# -*- coding: cp1252 -*-

# File:     TCAlgo_AlgoImplementation.py
# Project: 	The actual implementation of the algorithm.
# Author:	Axel Newe (axel.newe@fau.de)
# Version:	1.0
# Date: 	2015-11-11

# ------------------------------------------------------------------------------------------------------

from TCAlgo_AlgoSteps import *

# ------------------------------------------------------------------------------------------------------

def PerformAssessmentForAllNormalizedLabValueEpisodes(normalizedLabValueEpisodesDict, params):
    """
    This is the main processing loop.
    It takes ALL normalized lab values as input and returns a dictionary with ALL resulting assessments.
    :param normalizedLabValueEpisodesDict: A dictionary of medication days and lab values. Format: {'000': [[0, 1, 2, 3, 4, 5, 6], {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}]}
    :param params: The parameters tuple.
    :return: A dictionary that contains the assessments. Format: {'000': ['temporal correlation', 0.87], '001': ['no change', 0.99], '002': ['no assessment', 0.55]}
    """

    print "Assessing data..."

    # Init result
    assessmentDataDict = {}

    # Loop over all curves and call assessment function for each curve
    for thisCurveNumber in sorted(normalizedLabValueEpisodesDict.keys()):

        print "Processing curve #" + thisCurveNumber + "..."

        thisCurveAssessmentResult = PerformAssessmentForOneCurve(normalizedLabValueEpisodesDict[thisCurveNumber], params)

        print "Result: " + thisCurveAssessmentResult

        assessmentDataDict[thisCurveNumber] = thisCurveAssessmentResult

    print "Assessing data finished."

    return assessmentDataDict


def PerformMainAssessmentLoopForOneCurve(labValueEpisode, params, loopParams):
    """
    This is the main processing loop function for ONE curve episode. It performs several assessment steps using specified parameters.
    :param labValueEpisode: The lab value episode as a tuple of medication days and labe values, e.g. [[0, 1, 2, 3, 4, 5, 6], {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}]
    :param params: The parameter tuple.
    :param loopParams: The loop parameters tuple.
    :return: A string containing the assessment result, e.g. 'temporal correlation'
    """

    constantCurveThreshold, meanDifferenceThreshold, linearCurveThreshold, linearCurveAfterPrePhaseThreshold, meanDifferenceThreshold2 = loopParams

    # -----------------------------------------------------------------------------------------------
    # Main Loop Step 1: Check if lab value curve is constant during the pre-phase
    # (i.e., if values derive less than <constantCurveThreshold> from fitted constant curve)
    # but no longer in the mid-/post-phase (considering <meanDifferenceThreshold>).
    # -----------------------------------------------------------------------------------------------
    LabValueIsConstantInPrePhase = CheckForConstantCurve(labValueEpisode, constantCurveThreshold, PRE_PHASE)

    if (LabValueIsConstantInPrePhase):

        LabValuesDifferInMean = CheckIfMeanValueDiffersBetweenPhases(labValueEpisode, meanDifferenceThreshold)

        if (LabValuesDifferInMean):
            return ASSESSMENT_VALUE_TEMPORAL_CORRELATION


    # -----------------------------------------------------------------------------------------------
    # Main Loop Step 2: Check if lab value curve is linear
    # (i.e., if values derive less than <linearCurveThreshold> from fitted linear curve)
    # -----------------------------------------------------------------------------------------------

    LabValueCurveIsLinear = CheckForLinearCurve(labValueEpisode, linearCurveThreshold, ALL_PHASES)

    if (LabValueCurveIsLinear):
        return ASSESSMENT_VALUE_NO_CHANGE

    # -----------------------------------------------------------------------------------------------
    # Main Loop Step 3: Check if lab value curve is linear for mid-phase and post-phase
    # (i.e., if values derive less than <linearCurveAfterPrePhaseThreshold> from fitted linear curve).
    # -----------------------------------------------------------------------------------------------

    LabValueCurveIsLinearAfterPrePhase = CheckForLinearCurve(labValueEpisode, linearCurveAfterPrePhaseThreshold, MID_PHASE + POST_PHASE)

    if (LabValueCurveIsLinearAfterPrePhase):
        return ASSESSMENT_VALUE_NO_CHANGE

    # -----------------------------------------------------------------------------------------------
    # Main Loop Step 4: Check if mean values of pre-phase differ from mean values in mid-/post phase
    # (i.e., if pre-phase mean values derive more than <meanDifferenceThreshold2> from mean values of
    # later phases).
    # -----------------------------------------------------------------------------------------------

    LabValuesDifferInMean = CheckIfMeanValueDiffersBetweenPhases(labValueEpisode, meanDifferenceThreshold2)

    if (LabValuesDifferInMean):
        return ASSESSMENT_VALUE_TEMPORAL_CORRELATION

    # -----------------------------------------------------------------------------------------------
    # Main Loop Step 5: Return "not yet classified" if previous assessments yielded no result.
    # -----------------------------------------------------------------------------------------------
    return ASSESSMENT_VALUE_NOT_CLASSIFIED


def PerformAssessmentForOneCurve(normalizedLabValueEpisode, params):
    """
    This is the main processing function for ONE curve episode.
    :param normalizedLabValueEpisode: The lab value episode as a tuple of medication days and labe values, e.g. [[0, 1, 2, 3, 4, 5, 6], {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}]
    :param params: The parameter tuple.
    :return: A string containing the assessment result, e.g. 'temporal correlation'
    """

    # Unpack parameters
    noChangeIfInReferenceRange = params

    # -----------------------------------------------------------------------------------------------
    # Algorithm Preprocessing Step: Check if lab values are within reference range (only perform this
    # step if the appropriate switch has been set).
    # -----------------------------------------------------------------------------------------------
    if (noChangeIfInReferenceRange):

        LabValueCurveIsInReferenceRange = CheckIfValuesAreInReferenceRange(normalizedLabValueEpisode)

        if (LabValueCurveIsInReferenceRange):
            return ASSESSMENT_VALUE_NO_CHANGE

    # ********************************************************************************************
    # Main loop
    # ********************************************************************************************

    # ------------------------------------------------------------------------------------------------
    # Algorithm Step 1: Perform Main Loop with original data.
    # ------------------------------------------------------------------------------------------------
    constantCurveThreshold            = 0.125
    meanDifferenceThreshold           = 0.4
    linearCurveThreshold              = 0.125
    linearCurveAfterPrePhaseThreshold = 0.0625
    meanDifferenceThreshold2          = 0.5
    loopParams = (constantCurveThreshold, meanDifferenceThreshold, linearCurveThreshold, linearCurveAfterPrePhaseThreshold, meanDifferenceThreshold2)

    thisLoopResult = PerformMainAssessmentLoopForOneCurve(normalizedLabValueEpisode, params, loopParams)

    if (thisLoopResult != ASSESSMENT_VALUE_NOT_CLASSIFIED):
        return thisLoopResult

    # ------------------------------------------------------------------------------------------------
    # Algorithm Step 2: Apply low pass filter on lab values and use filtered data for Main Loop pass.
    # ------------------------------------------------------------------------------------------------
    normalizedLabValueEpisodeLowPass  = ApplyLowPassFilter(normalizedLabValueEpisode)
    constantCurveThreshold            = 0.05
    meanDifferenceThreshold           = 0.3
    linearCurveThreshold              = 0.1
    linearCurveAfterPrePhaseThreshold = 0.05
    meanDifferenceThreshold2          = 0.5
    loopParams = (constantCurveThreshold, meanDifferenceThreshold, linearCurveThreshold, linearCurveAfterPrePhaseThreshold, meanDifferenceThreshold2)

    thisLoopResult = PerformMainAssessmentLoopForOneCurve(normalizedLabValueEpisodeLowPass, params, loopParams)

    if (thisLoopResult != ASSESSMENT_VALUE_NOT_CLASSIFIED):
        return thisLoopResult

    # ------------------------------------------------------------------------------------------------
    # Algorithm Step 3: Remove outliers from original data, apply low pass filter and use this data
    # for next Main Loop pass.
    # ------------------------------------------------------------------------------------------------
    normalizedLabValueEpisodeWithoutOutliers = ReplaceOutliersByMedian(normalizedLabValueEpisode, 2)
    normalizedLabValueEpisodeLowPass2        = ApplyLowPassFilter(normalizedLabValueEpisodeWithoutOutliers)
    constantCurveThreshold                   = 0.025
    meanDifferenceThreshold                  = 0.2
    linearCurveThreshold                     = 0.025
    linearCurveAfterPrePhaseThreshold        = 0.025
    meanDifferenceThreshold2                 = 0.2
    loopParams = (constantCurveThreshold, meanDifferenceThreshold, linearCurveThreshold, linearCurveAfterPrePhaseThreshold, meanDifferenceThreshold2)

    thisLoopResult = PerformMainAssessmentLoopForOneCurve(normalizedLabValueEpisodeLowPass2, params, loopParams)

    if (thisLoopResult != ASSESSMENT_VALUE_NOT_CLASSIFIED):
        return thisLoopResult

    # ********************************************************************************************
    # Post-processing chain
    # ********************************************************************************************

    # --------------------------------------------------------------------------------------------
    # Algorithm Step 4: Check for differences in slope between the phases
    # --------------------------------------------------------------------------------------------

    LabValuesLowPassPrePhaseFallingOrRising = CheckIfLabValuesAreFallingOrRising(normalizedLabValueEpisodeLowPass, 0.1, PRE_PHASE)
    LabValuesLowPassPostPhaseFallingOrRising = CheckIfLabValuesAreFallingOrRising(normalizedLabValueEpisodeLowPass, 0.04, POST_PHASE)
    LabValuesLowPassMidPostPhaseFallingOrRising = CheckIfLabValuesAreFallingOrRising(normalizedLabValueEpisodeLowPass, 0.04, MID_PHASE + POST_PHASE)

    if (LabValuesLowPassPrePhaseFallingOrRising < 0): # Falling edge during pre-phase

        if (LabValuesLowPassMidPostPhaseFallingOrRising == 0) and (LabValuesLowPassPostPhaseFallingOrRising == 0):
            return ASSESSMENT_VALUE_NO_CHANGE
        else:
            return ASSESSMENT_VALUE_NO_ASSESSMENT

    elif (LabValuesLowPassPrePhaseFallingOrRising > 0): # Rising edge during pre-phase

        if (LabValuesLowPassMidPostPhaseFallingOrRising == 0) and (LabValuesLowPassPostPhaseFallingOrRising == 0):
            return ASSESSMENT_VALUE_NO_CHANGE
        else:
            return ASSESSMENT_VALUE_NO_ASSESSMENT

    # --------------------------------------------------------------------------------------------
    # Algorithm Step 5: Check for deviations from the mean value
    # --------------------------------------------------------------------------------------------

    LabValuesDifferFromMedian = CheckIfLabValuesDifferFromMean(normalizedLabValueEpisodeWithoutOutliers, 0.25)

    if (not LabValuesDifferFromMedian):
        return ASSESSMENT_VALUE_NO_CHANGE


    # --------------------------------------------------------------------------------------------
    # Algorithm Step 6: Check if lab values are alternating strongly (i.e., if they do "zig-zag").
    # --------------------------------------------------------------------------------------------

    LabValuesAlternateStrongly = CheckIfLabValuesAlternateStrongly(normalizedLabValueEpisode, 0.5, 5)
    if (LabValuesAlternateStrongly):
        return ASSESSMENT_VALUE_NO_ASSESSMENT

    # --------------------------------------------------------------------------------------------
    # Algorithm Last Step: Assume and return "temporal correlation" if everything else fails.
    # --------------------------------------------------------------------------------------------

    return ASSESSMENT_VALUE_TEMPORAL_CORRELATION
