# -*- coding: cp1252 -*-

# File:     TCAlgo_Tools.py
# Project:  General tool methods
# Author:	Axel Newe (axel.newe@fau.de)
# Version:	1.0
# Date: 	2015-11-11

# ------------------------------------------------------------------------------------------------------

import numpy as np
import scipy.ndimage as ndi
import math
import os.path
import sys

from TCAlgo_Defines import *

# ------------------------------------------------------------------------------------------------------

def GetLocalPath():
    """
    Returns the file system path where the software is executed.
    """
    return os.path.dirname(sys.argv[0])


def Signum(argument):
    """
    Returns the signum (i.e. +1 or -1) of the passed argmuent.
    :param argmuent: The argmunet for which the signum shall be calculated.
    :return: +1 or -1
    """

    return math.copysign(1, argument)


def GetUnbrokenArraysFromLabValuesDict(labValuesEpisode, consideredPhases, fillWithNaN = False):
    """
    Creates arrays for reference days and lab values without gaps (i.e., missing days).
    Missing values are filled up with a zero by default (or with a NaN by switch),
    but get a very high sigma value, so that it has only a negligible effect on curve fitting.
    :param labValuesEpisode: A list containing the list of reference days, e.g. [0, 1, 2, 3, 4, 5] and a dictionary of lab values, e.g. {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}
    :param consideredPhases: Specifies which phases (pre, mid, post) shall be considered.
    :param fillWithNaN: Specifies whether missing values shall be filled with NaN instead of 0 (Zero). Default = False
    :return: A list/array of reference days and lab values plus a sigma list with the uncertainties.
    """

    medicationDaysList, labValuesDict = labValuesEpisode
    lastMedicationDay = medicationDaysList[-1]

    considerPrePhase = ((consideredPhases & PRE_PHASE) == PRE_PHASE)
    considerMidPhase = ((consideredPhases & MID_PHASE) == MID_PHASE)
    considerPostPhase = ((consideredPhases & POST_PHASE) == POST_PHASE)

    ReferenceDayArray = []
    LabValuesArray    = []
    sigma             = []

    fillValue = 0

    if (fillWithNaN):
        fillValue = np.NaN

    SortedReferenceDaysFromDict = sorted(labValuesDict.keys())

    for thisReferenceDay in range(SortedReferenceDaysFromDict[0], SortedReferenceDaysFromDict[-1]+1):

        if  (thisReferenceDay < 0) and (not considerPrePhase):
            continue

        if  (thisReferenceDay >= 0) and (thisReferenceDay <= lastMedicationDay) and (not considerMidPhase):
            continue

        if  (thisReferenceDay > lastMedicationDay) and (not considerPostPhase):
            continue

        ReferenceDayArray += [thisReferenceDay]

        if thisReferenceDay in labValuesDict:
            LabValuesArray += [labValuesDict[thisReferenceDay]]
            sigma          += [SIGMA_CONSIDER]
        else:
            LabValuesArray += [fillValue]
            sigma          += [SIGMA_IGNORE]

    return np.array(ReferenceDayArray), np.array(LabValuesArray), np.array(sigma)


def RemoveDummyValues(valuesArray, sigmaArray):
    """
    Removes dummy values (which are values of SIGMA_CONSIDER, defined in TCAlgo_Defines.py) from array.
    :param valuesArray: A list containing the values.
    :param sigmaArray: A list containing the sigma values.
    :return: The updated values list.
    """

    result = []

    for i in range(len(valuesArray)):

        if (sigmaArray[i] == SIGMA_CONSIDER):
            result += [valuesArray[i]]

    return np.array(result)


def ApplyLowPassFilter(labValuesEpisode):
    """
    Applies a Gaussian low pass filter (sigma=1.5) to the lab values.
    :param labValuesEpisode: A list containing the list of reference days, e.g. [0, 1, 2, 3, 4, 5] and a dictionary of lab values, e.g. {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}
    :return: The same episode, but all lab values are low pass filtered.
    """

    labValueEpisodeLowPass = [labValuesEpisode[0]] # Add reference days from original data to result episode data

    labValuesDict = labValuesEpisode[1]
    labValuesDictLowPass = {}

    SortedReferenceDaysFromDict = sorted(labValuesDict.keys())

    temporaryLabValuesList = []

    for thisReferenceDay in SortedReferenceDaysFromDict:
        temporaryLabValuesList += [labValuesDict[thisReferenceDay]]

    temporaryLabValuesListLowPass = ndi.gaussian_filter1d(temporaryLabValuesList, sigma=1.5, output=np.float64, mode='nearest')

    for i in range(len(temporaryLabValuesListLowPass)):
        labValuesDictLowPass[SortedReferenceDaysFromDict[i]] = temporaryLabValuesListLowPass[i]

    labValueEpisodeLowPass += [labValuesDictLowPass] # Add low pass filtered lab values to result episode data

    return labValueEpisodeLowPass


def ReplaceOutliersByMedian(labValuesEpisode, maxNumOutliersToRemove = 2):
    """
    Removes outliers from data arrays and replaces them by the median. Assumes a linear curve.
    :param labValuesEpisode: A list containing the list of reference days, e.g. [0, 1, 2, 3, 4, 5] and a dictionary of lab values, e.g. {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}
    :param maxNumOutliersToRemove: The maximum number of outliers to remove.
    :return: The same episode, but with replaced outliers.
    """

    labValueEpisodeWithRemovedOutliers = [labValuesEpisode[0]] # Add reference days from original data to result episode data

    labValuesDict = labValuesEpisode[1]
    labValuesDictWithRemovedOutliers = {}

    SortedReferenceDaysFromDict = sorted(labValuesDict.keys())

    temporaryLabValuesList = []

    for thisReferenceDay in SortedReferenceDaysFromDict:
        temporaryLabValuesList += [labValuesDict[thisReferenceDay]]

    labValuesMedian = np.median(temporaryLabValuesList)

    temporaryLabValuesListWithRemovedOutliers = temporaryLabValuesList

    for i in range (maxNumOutliersToRemove):
        labValuesDeltaToMedian = np.abs(temporaryLabValuesListWithRemovedOutliers - labValuesMedian)
        maximumDeltaIndex = np.argmax(labValuesDeltaToMedian)
        temporaryLabValuesListWithRemovedOutliers[maximumDeltaIndex] = labValuesMedian

    for i in range(len(temporaryLabValuesListWithRemovedOutliers)):
        labValuesDictWithRemovedOutliers[SortedReferenceDaysFromDict[i]] = temporaryLabValuesListWithRemovedOutliers[i]

    labValueEpisodeWithRemovedOutliers += [labValuesDictWithRemovedOutliers] # Add low pass filtered lab values to result episode data

    return labValueEpisodeWithRemovedOutliers


def SplitLabValuesIntoPhases(labValuesEpisode):
    """
    Creates the array of lab values, one for each phase (before medication, during medication and after medication).
    :param labValuesEpisode: A list containing the list of reference days, e.g. [0, 1, 2, 3, 4, 5] and a dictionary of lab values, e.g. {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}
    :return: A tuple of three lab value np.arrays, one for each phase.
    """

    medicationDaysList, labValuesDict = labValuesEpisode

    labValuesPrePhase = []
    labValuesMedicationPhase = []
    labValuesPostPhase = []

    lastMedicationDay = medicationDaysList[-1]

    SortedReferenceDaysFromDict = sorted(labValuesDict.keys())

    for thisReferenceDay in SortedReferenceDaysFromDict:

        if (thisReferenceDay < 0):
            labValuesPrePhase += [labValuesDict[thisReferenceDay]]
        elif (thisReferenceDay > lastMedicationDay):
            labValuesPostPhase += [labValuesDict[thisReferenceDay]]
        else:
            labValuesMedicationPhase += [labValuesDict[thisReferenceDay]]


    return np.array(labValuesPrePhase), np.array(labValuesMedicationPhase), np.array(labValuesPostPhase)


def GetSortedLabValuesList(labValuesDict):
    """
    Extracts the lab values out of a lab values episode dictionary and sorts them in ascending order.
    :param labValuesDict: A dictionary of lab values, e.g. {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}
    :return: A list of these lab values, sorted by the key values of the dict in ascending order.
    """

    result = []

    referenceDays = sorted(labValuesDict.keys())

    for thisReferenceDay in referenceDays:
        result += [labValuesDict[thisReferenceDay]]

    return np.array(result)


def GetSlopesFromLabValuesDict(labValuesDict):
    """
    Extracts the slopes between two lab values out of a lab values episode dictionary.
    :param labValuesDict: A dictionary of lab values, e.g. {1: 0.25, 2: 0.167, 3: 0.111, 4: 0.222, 5: 0.361, 6: 0.333}
    :return: A list of these slopes.
    """

    SortedReferenceDaysFromDict = sorted(labValuesDict.keys())

    SlopesList = []
    NumLabValues = len(SortedReferenceDaysFromDict)

    for thisReferenceDayIndex in range(NumLabValues):
        if (thisReferenceDayIndex != (NumLabValues-1)):
            thisLabValue = labValuesDict[SortedReferenceDaysFromDict[thisReferenceDayIndex]]
            nextLabValue = labValuesDict[SortedReferenceDaysFromDict[thisReferenceDayIndex+1]]
            daysDelta = SortedReferenceDaysFromDict[thisReferenceDayIndex+1] - SortedReferenceDaysFromDict[thisReferenceDayIndex]
            thisSlope = float(nextLabValue - thisLabValue)/float(daysDelta)
            SlopesList += [thisSlope]

    return SlopesList
